#include "para.h"
#include "close_on_fork.h"

/*
 * fds must be a pointer to three integers, corresponding to stdin,
 * stdout and stderr respectively.
 * fd[i] < 0 : leave fd alone
 * fd[i] == 0: dup fd to /dev/null
 * fd[i] > 0 : dup fd and return fileno in fd[i]
 */
int para_exec(pid_t *pid, const char *file, char *const args[], int *fds)
{
	int in[2] = {-1, -1}, out[2] = {-1, -1}, err[2] = {-1, -1},
		null = -1; /* ;) */

	if (fds[0] > 0 && pipe(in) < 0)
		goto err_out;
	if (fds[1] > 0 && pipe(out) < 0)
		goto err_out;
	if (fds[2] > 0 && pipe(err) < 0)
		goto err_out;
	if (!fds[0] || !fds[1] || !fds[2]) {
		null = open("/dev/null", O_RDONLY);
		if (null < 0)
			goto err_out;
	}
	if ((*pid = fork()) < 0)
		exit(EXIT_FAILURE);
	if (!(*pid)) { /* child */
		close_listed_fds(); /* close unneeded fds */
		if (fds[0] >= 0) {
			if (fds[0]) {
				close(in[1]);
				if (in[0] != STDIN_FILENO)
					dup2(in[0], STDIN_FILENO);
			} else
				dup2(null, STDIN_FILENO);
		}
		if (fds[1] >= 0) {
			if (fds[1]) {
				close(out[0]);
				if (out[1] != STDOUT_FILENO)
					dup2(out[1], STDOUT_FILENO);
			} else
				dup2(null, STDOUT_FILENO);
		}
		if (fds[2] >= 0) {
			if (fds[2]) {
				close(err[0]);
				if (err[1] != STDERR_FILENO)
					dup2(err[1], STDERR_FILENO);
			} else
				dup2(null, STDERR_FILENO);
		}
		if (null >= 0)
			close(null);
		execvp(file, args);
		_exit(EXIT_FAILURE);
	}
	if (fds[0] > 0) {
		close(in[0]);
		*fds = in[1];
	}
	if (fds[1] > 0) {
		close(out[1]);
		*(fds + 1) = out[0];
	}
	if (fds[2] > 0) {
		close(err[1]);
		*(fds + 2) = err[0];
	}
	if (null >= 0)
		close(null);
	return 1;
err_out:
	if (err[0] >= 0)
		close(err[0]);
	if (err[1] >= 0)
		close(err[1]);
	if (out[0] >= 0)
		close(out[0]);
	if (out[1] >= 0)
		close(out[1]);
	if (in[0] >= 0)
		close(in[0]);
	if (in[1] >= 0)
		close(in[1]);
	if (null >= 0)
		close(null);
	return -1;
}


int para_exec_cmdline_pid(pid_t *pid, char *cmdline, int *fds)
{
	int argc, ret = -1;
	char **argv, *tmp = strdup(cmdline);

	if (!tmp)
		exit(EXIT_FAILURE);
	argc = split_args(tmp, &argv, ' ');
	if (argv) {
		ret = para_exec(pid, argv[0], argv, fds);
		free(argv);
	}
	free(tmp);
	return ret;
}

int file_exists(const char *fn)
{
        struct stat statbuf;

        return !stat(fn, &statbuf);
}
